package com.odin.pgpool.test;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.*;
import java.util.Random;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;


public class Issue271 {

    private static final int TEST_THREADS_NUM = 60;
    static volatile boolean writePerformed = false; // variable will be set into TRUE when the write thread did INSERT operation

    public static void main(String[] args) throws Exception {
        java.sql.DriverManager.setLogWriter(new java.io.PrintWriter(System.out));
        Random random = new Random(System.currentTimeMillis());

        try (Connection conn = DriverManager.getConnection("jdbc:postgresql://10.28.78.115/oss?loglevel=2", "oss", "H76iOavIst0EJ3")) {

            Lock lock = new ReentrantLock();

            conn.setAutoCommit(false); // here we define explicit transaction
            CyclicBarrier cb = new CyclicBarrier(TEST_THREADS_NUM);

            final int valueTest = random.nextInt();
            System.out.println("test value:" + valueTest);

            try (PreparedStatement stmtSelect = conn.prepareStatement("select * from write_table where id = ?");
                 PreparedStatement stmtWrite = conn.prepareStatement("insert into write_table(id) values(?)")) {

                List<Thread> thrs = new ArrayList<>();
                for (int i = 0; i < TEST_THREADS_NUM; i++) {
                    boolean isRead = i != 11;
                    //boolean isSleep = i < (TEST_THREADS_NUM / 2);
                    thrs.add(new TestReader(valueTest, cb, isRead, isRead ? stmtSelect : stmtWrite, lock));
                }

                for (Thread thr : thrs) {
                    thr.start();
                }

                Thread.sleep(1000);

                for (Thread thr : thrs) {
                    thr.join();
                }


            } catch (Exception ex) {
                ex.printStackTrace();

                throw ex;
            }
            conn.commit();
        }
    }

    public static class TestReader extends Thread {

        private final CyclicBarrier cb;

        private final PreparedStatement stmt;
        private final Lock lock;

        private final boolean isRead;
        private final int valueTest;

        TestReader(int valueTest, CyclicBarrier cb, boolean readThread, PreparedStatement stmt, Lock lock) {
            this.valueTest = valueTest;
            this.cb = cb;
            this.isRead = readThread;
            this.stmt = stmt;
            this.lock = lock;
        }

        @Override
        public void run() {
            try {
                cb.await();

                lock.lock();
                stmt.setInt(1, valueTest);
                if (isRead) {
                    try (ResultSet rs = stmt.executeQuery()) {
                        if (writePerformed) {
                            if (!rs.next()) {
                                throw new RuntimeException("The request was balanced. The value was inserted but the the following select returned nothing");
                            } else {
                                int val = rs.getInt(1);
                                System.out.println("recevied value:" + val);
                            }
                        } else {
                            // write was not performed so one should not expect the value is in the database
                        }
                    }
                } else {
                    stmt.execute();
                    System.out.println("WRITE DOWN");
                    writePerformed = true; // set mark the write performed. since this moment the READ threads will be trying to read the value.
                }

            } catch (Exception ex) {
                ex.printStackTrace();
                throw new RuntimeException(ex);
            } finally {
                lock.unlock();
            }

        }

    }
}
